/* ez-blob-io.c: Common i/o utilities for serialisers

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include "ez-blob.h"
#include "ez-blob-io.h"

//#define ABORT() abort()
#define ABORT()

void *blobio_reserve(struct ez_blobio * __restrict io, size_t len) {
	size_t size = io->size;
	size_t to = io->index + len;
	void *v;

	if (io->error)
		return NULL;

	switch (io->mode) {
	case BLOBIO_WRITE_ALLOC:
		if (to > size) {
			void *alloc;
			
			do {
				size = size ? size * 2 : 256;
			} while (to > size);
			
			io->size = size;
			alloc = realloc(io->data, size);

			if (!alloc) {
				io->error = ENOMEM;
				printf("realloc failed\n");
				ABORT();
				return NULL;
			}

			io->data = alloc;
		}

		v = io->index + io->data;
		io->index = to;
		return v;

	case BLOBIO_WRITE_FIXED:
		if (to > size) {
			io->error = EOVERFLOW;
			ABORT();
			return NULL;
		}
		v = io->index + io->data;
		io->index = to;
		return v;

	case BLOBIO_WRITE_SIZE:
		io->index = to;
		return NULL;
	default:
		io->error = EINVAL;
		return NULL;
	}
}

void *blobio_take(struct ez_blobio * __restrict io, size_t len) {
	size_t to = io->index + len;

	if (to <= io->size) {
		void *v = io->data + io->index;
		io->index = to;
		return v;
	} else {
		ABORT();
		io->index = io->size;
		io->error = EOVERFLOW;
		return NULL;
	}
}

void blobio_write_align(struct ez_blobio *io, unsigned int step) {
	int skip = (step - io->index) & (step-1);
	uint8_t *v;

	if (skip && (v = blobio_reserve(io, skip))) {
		for (int i=0;i<skip;i++)
			v[i] = 0;
	}
}

void blobio_read_align(struct ez_blobio *io, unsigned int step) {
	int skip = (step - io->index) & (step-1);

	blobio_take(io, skip);
}
