/* ez-blob-xdrn.c: XDR-like serialiser using native ordering.

   Copyright (C) 2019 Michael Zucchi

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation, either version 3 of
   the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this program. If not, see
   <http://www.gnu.org/licenses/>.
*/

/*
  Implements a XDR serialiser using native byte ordering.

  Each end must have identical descriptor tables.

  Support matrix:

	       	     single  singlep vector  vectorp list    listp   cstring
                             ?null 1                         ?null   ?null

      int8  	        x      	       	x 2   	       	       	       	x 3
      int16 	        x	 	x 2
      int32 	        x	 	x
      int64 	        x	 	x
      float 	        x	 	x
      double	        x	 	x
      struct	        x      	x      	       	       	x     	x

  1. Nullable fields are optional and must be properly annotated.
     They are written as Optional-Data.
  2. These aren't XDR compatible, it's written as a Variable-Length Array
     but with 2-byte elements and end-padding.  Well int8 is identical
     to Opaque Data.
  3. These are 8-bit strings (e.g. UTF-8), not necessarily ASCII as
     specified by String.

*/

#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include <errno.h>

#include "ez-list.h"

#include "ez-blob.h"
#include "ez-blob-xdrn.h"
#include "ez-blob-io.h"

static size_t roundup(size_t v) __attribute__ ((always_inline));
static __inline__ size_t roundup(size_t v) {
	return (v+3) & ~3;
}

/*
  Arrays of byte and short are encoded as if they were Variable-Length
  Array but with natural sized elements, followed by padding.
*/

static void xdrio_writev(struct ez_blobio *io, const ez_blob *data, size_t elshift) {
	size_t size = data->eb_size << elshift;
	int32_t count = data->eb_size;
	void *v;

	blobio_write32(io, count);
	v = blobio_reserve(io, size);
	if (v) {
		memcpy(v, data->eb_data, size);
		blobio_write_align(io, 4);
	}
}

static void xdrio_readv(struct ez_blobio *io, ez_blob *data, size_t elshift) {
	int32_t count = blobio_readi32(io);
	size_t size = count << elshift;
	void *src = blobio_take(io, size);

	if (src) {
		void *mem = malloc(size);

		if (mem) {
			memcpy(mem, src, size);
			data->eb_data = mem;
			data->eb_size = count;
			blobio_read_align(io, 4);
		} else {
			io->error = ENOMEM;
		}
	}
}

static void xdrio_writes(struct ez_blobio *io, const char *v) {
	if (v) {
		int32_t size = strlen(v);

		blobio_write32(io, size);
		blobio_write(io, v, size);
		blobio_write_align(io, 4);
	} else {
		io->error = EINVAL;
	}
}

static void xdrio_reads(struct ez_blobio *io, char **v) {
	uint32_t size = blobio_readi32(io);
	void *src = blobio_take(io, size);

	if (src) {
		char *mem = malloc(size+1);

		if (mem) {
			memcpy(mem, src, size);
			mem[size] = 0;
			*v = mem;
			blobio_read_align(io, 4);
		} else {
			io->error = ENOMEM;
		}
	}
}

static void xdrn_encode_raw(struct ez_blobio *io, const ez_blob_desc *desc, const void *p) {
	for (int i=0,dlen=desc->bd_length;i<dlen;i++) {
		const ez_blob_desc *d = &desc[i+1];
		const void *v = (p + d->bd_offset);

		switch (d->bd_type) {
			/*
			 * Primitive x 1
			 */
		case EZ_BLOB_SINGLE | EZ_BLOB_INT8:
			blobio_write32(io, *(uint8_t *)v);
			break;
		case EZ_BLOB_SINGLE | EZ_BLOB_INT16:
			blobio_write32(io, *(uint16_t *)v);
			break;
		case EZ_BLOB_SINGLE | EZ_BLOB_INT32:
			blobio_write32(io, *(uint32_t *)v);
			break;
		case EZ_BLOB_SINGLE | EZ_BLOB_INT64:
			blobio_write64(io, *(uint64_t *)v);
			break;
		case EZ_BLOB_SINGLE | EZ_BLOB_FLOAT32:
			blobio_writef(io, *(float *)v);
			break;
		case EZ_BLOB_SINGLE | EZ_BLOB_FLOAT64:
			blobio_writed(io, *(double *)v);
			break;

			/*
			 * Primitive x array
			 */
		case EZ_BLOB_VECTOR | EZ_BLOB_INT8:
		case EZ_BLOB_VECTOR | EZ_BLOB_INT16:
		case EZ_BLOB_VECTOR | EZ_BLOB_INT32:
		case EZ_BLOB_VECTOR | EZ_BLOB_INT64:
		case EZ_BLOB_VECTOR | EZ_BLOB_FLOAT32:
		case EZ_BLOB_VECTOR | EZ_BLOB_FLOAT64:
			xdrio_writev(io, (const ez_blob *)v, (d->bd_type & EZ_BLOB_SIZE));
			break;

			/*
			 * C string byte
			 */
		case EZ_BLOB_CSTRING | EZ_BLOB_INT8 | EZ_BLOB_ISNULLABLE:
			if (!*(const char **)v) {
				blobio_write32(io, 0);
				break;
			}
			blobio_write32(io, 1);
			// falls through
		case EZ_BLOB_CSTRING | EZ_BLOB_INT8:
			xdrio_writes(io, *(const char **)v);
			break;

			/*
			 * Structure x 1
			 */
		case EZ_BLOB_STRUCT | EZ_BLOB_SINGLEP | EZ_BLOB_ISNULLABLE:
			if (!*(const void **)v) {
				blobio_write32(io, 0);
				break;
			}
			blobio_write32(io, 1);
			// falls through
		case EZ_BLOB_STRUCT | EZ_BLOB_SINGLEP:
			v = *(const void **)v;
			if (!v) {
				io->error = EINVAL;
				return;
			}
			// falls through
		case EZ_BLOB_STRUCT | EZ_BLOB_SINGLE:
			xdrn_encode_raw(io, d->bd_table, v);
			break;

			/*
			 * Structure x list
			 */
		case EZ_BLOB_STRUCT | EZ_BLOB_LISTP | EZ_BLOB_ISNULLABLE:
			if (!*(const void **)v) {
				blobio_write32(io, 0);
				break;
			}
			blobio_write32(io, 1);
			// falls through
		case EZ_BLOB_STRUCT | EZ_BLOB_LISTP:
			v = *(const void **)v;
			if (!v) {
				io->error = EINVAL;
				return;
			}
			// falls through
		case EZ_BLOB_STRUCT | EZ_BLOB_LIST:
			blobio_write32(io, ez_list_size((ez_list *)v));
			for (ez_node *w = ez_list_head((ez_list *)v), *n = ez_node_succ(w); n; w = n, n=ez_node_succ(n)) {
				xdrn_encode_raw(io, d->bd_table, w);
			}
			break;
		default:
			io->error = ENOTSUP;
			return;
		}
	}
}

size_t ez_xdrn_size(const ez_blob_desc *d, const void *p) {
	struct ez_blobio io = {
		.mode = BLOBIO_WRITE_SIZE
	};
	xdrn_encode_raw(&io, d, p);

	return io.index;
}

int ez_xdrn_encode_raw(const ez_blob_desc *d, const void *p, ez_blob *blob) {
	struct ez_blobio io = {
		.data = blob->eb_data,
		.size = blob->eb_size,
		.mode = BLOBIO_WRITE_FIXED
	};

	xdrn_encode_raw(&io, d, p);

	return io.error;
}

int ez_xdrn_encode(const ez_blob_desc *d, const void *p, ez_blob *blob) {
	struct ez_blobio io = {
		.mode = BLOBIO_WRITE_ALLOC
	};

	xdrn_encode_raw(&io, d, p);

	if (!io.error) {
		blob->eb_size = io.index;
		blob->eb_data = io.data;
		return 0;
	}

	free(io.data);
	blob->eb_data = NULL;
	blob->eb_size = 0;

	return io.error;
}

static void xdrn_decode_raw(struct ez_blobio *io, const ez_blob_desc *desc, void *p) {
	for (int i=0,dlen=desc->bd_length;i<dlen;i++) {
		const ez_blob_desc *d = &desc[i+1];
		void *v = (p + d->bd_offset);

		switch (d->bd_type) {
			/*
			 * Primitive x 1
			 */
		case EZ_BLOB_SINGLE | EZ_BLOB_INT8:
			*(uint8_t *)v = blobio_readi32(io);
			break;
		case EZ_BLOB_SINGLE | EZ_BLOB_INT16:
			*(uint16_t *)v = blobio_readi32(io);
			break;
		case EZ_BLOB_SINGLE | EZ_BLOB_INT32:
			*(uint32_t *)v = blobio_readi32(io);
			break;
		case EZ_BLOB_SINGLE | EZ_BLOB_INT64:
			*(uint64_t *)v = blobio_readi64(io);
			break;
		case EZ_BLOB_SINGLE | EZ_BLOB_FLOAT32:
			*(float *)v = blobio_readf(io);
			break;
		case EZ_BLOB_SINGLE | EZ_BLOB_FLOAT64:
			*(double *)v = blobio_readd(io);
			break;

			/*
			 * Primitive x array
			 */
		case EZ_BLOB_VECTOR | EZ_BLOB_INT8:
		case EZ_BLOB_VECTOR | EZ_BLOB_INT16:
		case EZ_BLOB_VECTOR | EZ_BLOB_INT32:
		case EZ_BLOB_VECTOR | EZ_BLOB_INT64:
		case EZ_BLOB_VECTOR | EZ_BLOB_FLOAT32:
		case EZ_BLOB_VECTOR | EZ_BLOB_FLOAT64:
			xdrio_readv(io, (ez_blob *)v, (d->bd_type & EZ_BLOB_SIZE));
			break;

			/*
			 * C string byte
			 */
		case EZ_BLOB_CSTRING | EZ_BLOB_INT8 | EZ_BLOB_ISNULLABLE:
			if (blobio_readi32(io) == 0)
				break;
			// falls through
		case EZ_BLOB_CSTRING | EZ_BLOB_INT8:
			xdrio_reads(io, (char **)v);
			break;

			/*
			 * Structure x 1
			 */
		case EZ_BLOB_STRUCT | EZ_BLOB_SINGLEP | EZ_BLOB_ISNULLABLE:
			if (blobio_readi32(io) == 0)
				break;
			// falls through
		case EZ_BLOB_STRUCT | EZ_BLOB_SINGLEP:
			*(void **)v = ez_blob_alloc(d->bd_table);
			v = *(void **)v;
			if (!v) {
				io->error = ENOMEM;
				return;
			}
			// falls through
		case EZ_BLOB_STRUCT | EZ_BLOB_SINGLE:
			xdrn_decode_raw(io, d->bd_table, v);
			break;

			/*
			 * Structure x list
			 */
		case EZ_BLOB_STRUCT | EZ_BLOB_LISTP | EZ_BLOB_ISNULLABLE:
			if (blobio_readi32(io) == 0)
				break;
			// falls through
		case EZ_BLOB_STRUCT | EZ_BLOB_LISTP:
			*(ez_list **)v = malloc(sizeof(ez_list));
			v = *(void **)v;
			if (!v) {
				io->error = ENOMEM;
				return;
			}
			ez_list_init((ez_list *)v);
			// falls through
		case EZ_BLOB_STRUCT | EZ_BLOB_LIST: {
			uint32_t count = blobio_readi32(io);

			for (int j=0;j<count;j++) {
				void *node = ez_blob_alloc(d->bd_table);

				if (node) {
					xdrn_decode_raw(io, d->bd_table, node);
					ez_list_addtail((ez_list *)v, node);
				} else {
					io->error = ENOMEM;
					return;
				}
			}
			break;
		}
		default:
			io->error = EILSEQ;
			return;
		}
	}
}

int ez_xdrn_decode_raw(const ez_blob_desc *desc, const ez_blob *blob, void *p) {
	struct ez_blobio io = {
		.size = blob->eb_size,
		.data = blob->eb_data,
		.mode = BLOBIO_READ
	};

	ez_blob_init(desc, p);
	xdrn_decode_raw(&io, desc, p);

	if (!io.error)
		return 0;
	
	ez_blob_free_raw(desc, p);
	return io.error;
}

void *ez_xdrn_decode(const ez_blob_desc *desc, const ez_blob *blob) {
	void *p = malloc(desc->bd_offset);

	if (p && ez_xdrn_decode_raw(desc, blob, p) != 0) {
		free(p);
		p = NULL;
	}

	return p;
}
